/* Copyright 2009, UCAR/Unidata and OPeNDAP, Inc.
   See the COPYRIGHT file for more information. */

#include "ocinternal.h"
#include "ocdebug.h"

#define LBRACKET '['
#define RBRACKET ']'


/* Forward reference */
static int decodeparameters(DAPURL*,char*);

static char* legalprotocols[] = {
"file:",
"http:",
"https:",
"ftp:",
NULL /* NULL terminate*/
};


/* Do a simple url parse*/
int
dapurlparse(const char* url0, DAPURL* dapurl)
{
    char* url;
    char** pp;
    char* p;
    char* p1;
    int c;

    /* accumulate parse points*/
    char* protocol = NULL;
    char* params = NULL;
    char* baseurl = NULL;
    char* constraints = NULL;
    char* stop;

    memset((void*)dapurl,0,sizeof(DAPURL));

    /* copy url and remove all whitespace*/
    url = strdup(url0);

    p = url;
    p1 = url;
    while((c=*p1++)) {if(c != ' ' && c != '\t') *p++ = c;}

    p = url;
    stop = p + strlen(p);

    /* break up the url string into pieces*/
    if(*p == LBRACKET) {
	params = p+1; /* leave off the leading lbracket*/
	/* find end of the clientparams*/
        for(;*p;p++) {if(p[0] == RBRACKET && p[1] != LBRACKET) break;}
	if(*p == 0) goto fail; /* malformed client params*/
	*p = '\0'; /* leave off the trailing rbracket	*/
	p++; /* move past the params*/
    }

    /* verify that the url starts with an acceptable protocol*/
    for(pp=legalprotocols;*pp;pp++) {
        if(strncmp(p,*pp,strlen(*pp))==0) break;
    }
    if(*pp == NULL) goto fail; /* illegal protocol*/
    /* save the protocol */
    protocol = *pp;

    baseurl = p;

    /* Look for '?' */
    constraints = strchr(p,'?');
    if(constraints) {
	*constraints = '\0';
    }

    /* assemble the component pieces*/
    dapurl->url = strdup(url0);
    dapurl->base = strdup(baseurl);
    dapurl->protocol = strdup(protocol);
    /* remove trailing : */
    dapurl->protocol[strlen(protocol)-1] = '\0';
    dapurlsetconstraints(dapurl,constraints);

    /* parse the parameters, if any*/
    if(params != NULL) {
	if(!decodeparameters(dapurl,params)) goto fail;
    }


    if(ocdebug) {
        fprintf(stderr,"dapurl: base=|%s| projection=|%s| selection=|%s|\n",
		dapurl->base, dapurl->projection, dapurl->selection);

    }
    free(url);
    return 1;

fail:
    if(url != NULL) free(url);
    return 0;
}

static int
decodeparameters(DAPURL* dapurl, char* params0)
{
    char* cp;
    char* cq;
    char** ep;
    int c;
    int i;
    int nparams;
    char** envv = NULL;
    char* params = strdup(params0);
    char* params1 = strdup(params);

    /* Pass 1 to replace "][" pairs with ','*/
    cp=params; cq = params1;
    while((c=*cp++)) {
	if(c == RBRACKET && *cp == LBRACKET) {cp++; c = ',';}
	*cq++ = c;
    }
    *cq = '\0';
    free(params);
    params = params1;

    /* Pass 2 to count # of params and break string*/
    nparams = 0;
    for(cp=params;(c=*cp);cp++) {
	if(c == ',') {*cp = '\0'; nparams++;}
    }
    nparams++; /* Add one for the last parameter*/
    envv = (char**)ocmalloc(sizeof(char*)*((2*nparams)+1));
    MEMCHECK(envv,0);

    /* Pass 3 to break up each pass into a (name,value) pair*/
    /* and insert into the envv*/
    /* parameters of the form name name= are converted to name=""*/
    cp = params;
    ep = envv;
    for(i=0;i<nparams;i++) {
	char* next = cp+strlen(cp)+1; /* save ptr to next pair*/
	char* vp;
	/*break up the ith param*/
	vp = strchr(cp,'=');
	if(vp != NULL) {*vp = '\0'; vp++;} else {vp = "";}
        *ep++ = strdup(cp);
	*ep++ = strdup(vp);
	cp = next;
    }
    *ep = NULL; /* null terminate*/
    free(params);
    dapurl->params = envv;
    return 1;
}

/* Call must free the actual url instance.*/
void
dapurlclear(DAPURL* dapurl)
{
    if(dapurl->url != NULL) {free(dapurl->url);}
    if(dapurl->base != NULL) {free(dapurl->base);}
    if(dapurl->protocol != NULL) {free(dapurl->protocol);}
    if(dapurl->projection != NULL) {free(dapurl->projection);}
    if(dapurl->selection != NULL) {free(dapurl->selection);}
    if(dapurl->params != NULL) {
	char** pp;
	for(pp=dapurl->params;*pp;pp+=2) {
	    free(pp[0]);
	    if(pp[1] != NULL) free(pp[1]);
        }
	free(dapurl->params);
    }
    memset((void*)dapurl,0,sizeof(DAPURL));
}

const char*
dapurllookup(DAPURL* durl, const char* clientparam)
{
    char** params = durl->params;
    if(params == NULL || clientparam == NULL) return NULL;
    while(*params) {
	char* name = *params++;
	char* value = *params++;
	if(strcmp(clientparam,name)==0) return value;
    }
    return NULL;
}

/*
If value is NULL, then param is removed.
return value = 1 => found and inserted/replaced/deleted;
               0 => param not found
*/
int
dapurlreplace(DAPURL* durl, const char* clientparam, const char* newvalue)
{
    char** params = durl->params;
    unsigned int nstrings;
    char** location;

    if(clientparam == NULL) return 0;
    /* simultaneously count and search*/
    location = NULL;
    nstrings = 0;
    if(params != NULL) {
        while(*params) {
	    char* name = *params;
	    if(strcmp(clientparam,name)==0) {location = params;}
	    nstrings++;
	    params += 2;
	}
    }
    if(location == NULL && newvalue == NULL)
	return 0; /* not found and not inserting */
    if(location == NULL && newvalue != NULL) { /* insert */
	size_t oldlen = (2*(nstrings)); /* excluding null terminator */
	size_t newsize = (oldlen+2+1)*sizeof(char*); /* include terminator */
        char** newparams = (char**)ocmalloc(newsize);
        if(newparams == NULL) return 0;
	if(durl->params != NULL) {
	    memcpy((void*)newparams,(void*)durl->params,oldlen*sizeof(char*));
	}	
	newparams[oldlen+0] = strdup(clientparam);
	newparams[oldlen+1] = strdup(newvalue);
	newparams[oldlen+2] = NULL;
	ocfree(durl->params);
	durl->params = newparams;
	return 1;
    } else if(location != NULL && newvalue != NULL) { /* replace*/
	ocfree(location[1]);
	location[1] = strdup(newvalue);
    } else { /* delete*/
	unsigned int count, rem;
	count = (nstrings*2)+1; /* number of char* instances in durl->params*/
	rem = ((durl->params+count) - (location+2));
	ocfree(location[0]);
	ocfree(location[1]);
	memcpy((void*)location,(void*)(location+2),rem*sizeof(char*));
    }
    return 1;
}

/* Replace the constraints */
void
dapurlsetconstraints(DAPURL* durl,const char* constraints)
{
    char* proj = NULL;
    char* select = NULL;
    const char* p;

    if(durl->projection != NULL) ocfree(durl->projection);
    if(durl->selection != NULL) ocfree(durl->selection);
    durl->projection = NULL;	
    durl->selection = NULL;

    if(constraints == NULL) return;

    p = constraints;
    if(p[0] == '?') p++;
    proj = (char*) p;
    select = strchr(proj,'&');
    if(select != NULL) {
        size_t plen = (select - proj);
	if(plen == 0) {
	    proj = NULL;
	} else {
	    proj = (char*)ocmalloc(plen+1);
	    memcpy((void*)proj,proj,plen);
	    proj[plen] = '\0';
	}
	select = strdup(select);
    } else {
	proj = strdup(proj);
	select = NULL;
    }
    durl->projection = proj;
    durl->selection = select;
}

/* Construct a complete DAP URL without the client params
   and optionally with the constraints;
   caller frees returned string
*/
char*
dapurlgeturl(DAPURL* durl, const char* prefix, const char* suffix,
             int withconstraints)
{
    size_t len = 0;
    char* newurl;

    len += strlen(durl->base);
    if(prefix != NULL) len += strlen(prefix);
    if(suffix != NULL) len += strlen(suffix);
    if(withconstraints
	&& (durl->projection != NULL || durl->selection != NULL)) {
	len += 1; /* leading '?' */
	if(durl->projection != NULL) len += strlen(durl->projection);
	if(durl->selection != NULL) len += strlen(durl->selection);
    }
    len += 1; /* null terminator */
    
    newurl = (char*)ocmalloc(len);
    if(!newurl) return NULL;
    newurl[0] = '\0';
    if(prefix != NULL) strcat(newurl,prefix);
    strcat(newurl,durl->base);
    if(suffix != NULL) strcat(newurl,suffix);
    if(withconstraints
	&& (durl->projection != NULL || durl->selection != NULL)) {
	strcat(newurl,"?");
	if(durl->projection != NULL) strcat(newurl,durl->projection);
	if(durl->selection != NULL) strcat(newurl,durl->selection);
    }
    return newurl;
}
