/* Copyright 2009, UCAR/Unidata and OPeNDAP, Inc.
   See the COPYRIGHT file for more information. */

#include "ocinternal.h"
#include "ocdata.h"
#include "occontent.h"
#include "i.h"


int verbose;
char* fileurl;
int debug;
int metadata;
extern int ocdebug;

extern int optind;
extern int opterr;
extern char* optarg;

/* Forward*/
static void usage(void);
static int fail(void);
static void check_err(int stat);
static void dumpflags(void);
static char* indent(int);

static int processdata(char*);
static void printdds(OCnode* inode, int depth);
static void printleaves(OCnode*);
static void printleaves1(OCnode*, OCcontent* content, OCtype octype);
static void dumpinstance(char* name, int rank, int* dims, size_t count, OCtype, char* memory);

int
main(int argc, char **argv)
{
    int c;
    int ncid;

    debug = 0;
    verbose=0;
    metadata=0;
    fileurl = NULL;

    opterr=1;
    while ((c = getopt(argc, argv, "dvau:D:l:p:c:t:R:")) != EOF) {
      switch(c) {
      case 'v': verbose=1; break;
      case 'd': debug=1; break;
      case 'D': debug=atoi(optarg); break;
      case 'm': metadata=1; break;
      default: break;
      }
    }

    if(debug > 0) {ocdebug = debug;}

    argc -= optind;
    argv += optind;

    if(argc > 0 && fileurl == NULL) {
	fileurl = strdup(argv[argc-1]);
    }

    if(fileurl == NULL) fileurl = getenv("FILEURL");

    if(fileurl == NULL) {
	fprintf(stderr,"no file url specified\n");
	usage();
    }

    if(verbose) dumpflags();

    if(verbose) {fprintf(stdout,"initializing\n"); fflush(stdout);}

    processdata(fileurl);

    return 0;
}

static void
dumpflags(void)
{
    fprintf(stderr,"url=%s\n",fileurl);
    if(verbose) fprintf(stderr,"verbose=%d\n",verbose);
    if(debug) fprintf(stderr,"debug=%d\n",debug);
}

static void
usage(void)
{
    fprintf(stderr,"usage: test [-t] [-d] [-oc] [-nd] [-r] [-u] <fileurl>\n");
    fail();
}

static void
check_err(int stat)
{
    if(stat == OC_NOERR) return;
    fprintf(stderr,"error status returned: (%d) %s\n",stat,ocerrstring(stat));
    fail();
}

static int
fail(void)
{
    fflush(stdout); fflush(stderr);
    exit(1);
}

static char*
indent(int depth)
{
    int size = (depth*2);
    char* blanks = malloc(size+1);
    memset(blanks,' ',size);
    blanks[size] = '\0';
    return blanks;
}

/*********************************/

static int
processdata(char* fileurl)
{
    OCnode* inode;
    int stat;

    stat = opendap_connect(fileurl, "", &inode, 0, 0);
    check_err(stat);

    if(metadata) {
        printdds(inode,0);
    } else {
        printdds(inode,0);
        printleaves(inode);
    }
}

static void
printdds(OCnode* inode, int depth)
{
    char* name;
    OCtype octype;
    int nsubnodes;
    int hascontent;
    int hasattrs;
    OCcontent* content;
    int i,stat;
    int* dimdata;
    char* aname;
    char** dimnames;
    int rank;
    OCtype primtype;

    stat = opendap_node_info(inode,&name,&octype,&nsubnodes,&hascontent,&hasattrs,&content);
    check_err(stat);

    if(octype == OC_Primitive) {
	/* get dimension info*/
	stat = opendap_get_array(content,(void*)&dimdata,&aname,&dimnames,0);
	check_err(stat);
	/* dump dim info*/
	rank = dimdata[0];
	primtype = (OCtype)dimdata[rank+1];
        fprintf(stdout,"[%2d]: %s %s", depth,octypetostring(primtype),name);
	for(i=0;i<rank;i++) {
	    if(strcmp(dimnames[i]," ")==0)
	        fprintf(stdout,"[%d]",dimdata[1+i]);
	    else
	    fprintf(stdout,"[%s=%d]",dimnames[i],dimdata[1+i]);
	}
	fprintf(stdout,"\n");
    } else if(octype == OC_Sequence) {
	int numrows,numfields;
        stat = opendap_get_sequence_info(content, &numrows, &numfields);
	check_err(stat);
        fprintf(stdout,"[%2d]: %s %s |columns|=%d |rows|=%d\n",
		depth,octypetostring(octype),name,numfields,numrows);
	for(i=0;i<nsubnodes;i++) {
	    OCnode* subnode;
	    stat = opendap_get_subnode(inode,i,&subnode,&name);
	    printdds(subnode,depth+1);
	}
    } else {
        fprintf(stdout,"[%2d]: %s %s |subnodes|=%d\n",
		depth,octypetostring(octype),name,nsubnodes);
	for(i=0;i<nsubnodes;i++) {
	    OCnode* subnode;
	    stat = opendap_get_subnode(inode,i,&subnode,&name);
	    printdds(subnode,depth+1);
	}
    }
}

static void
printleaves(OCnode* node)
{
    char* name;
    OCtype octype;
    int nsubnodes;
    int hascontent;
    int hasattrs;
    OCcontent* content;
    int i,stat;

    stat = opendap_node_info(node,&name,&octype,&nsubnodes,&hascontent,&hasattrs,&content);
    check_err(stat);

    /* walk the content*/
    printleaves1(node, content, octype);
}

static void
printleaves1(OCnode* inode, OCcontent* content, OCtype octype)
{
    int stat;
    int nmembers;
    char* aname;
    char** dimnames;
    int* dimdata;
    char* memory;
    int i,rank,nrows;
    size_t count;

    switch (octype) {
    case OC_Structure:
    case OC_Dataset:
	stat = opendap_get_struct_info(content,&nmembers);
	check_err(stat);
        for(i=0;i<nmembers;i++) {
	    OCcontent* sub;
	    OCtype subtype;
	    OCnode* subnode;
	    char* name;

	    stat = opendap_get_subnode(inode, i, &subnode, &name);
	    check_err(stat);
	    free(name); /* not needed*/
	    stat = opendap_get_struct_member(content,i,&subtype,&sub);
	    check_err(stat);
	    printleaves1(subnode,sub,subtype);
	}
	break;

    case OC_Primitive: /* leaf*/
	/* Get the dimensional info*/
	stat = opendap_get_array(content,(void*)&dimdata,&aname,&dimnames,0);
	check_err(stat);
	rank = dimdata[0];
	octype = (OCtype)dimdata[rank+1];
	count = dimdata[rank+2];
	/* get the actual data*/
	stat = opendap_get_array(content,(void*)&memory,&aname,&dimnames,1);
	check_err(stat);
	dumpinstance(aname,rank,dimdata+1,count,octype,memory);
	break;

    case OC_Sequence:
	stat = opendap_get_sequence_info(content,&nrows,&nmembers);
	check_err(stat);
        for(i=0;i<nmembers;i++) {
	    OCcontent* sub;
	    OCnode* subnode;
	    char* name;
	    OCtype subtype;
	    OCcontent* rowcolcontent;
	    int j;

	    stat = opendap_get_sequence_field(inode,content,i,&name,
					      &subtype,&subnode,&sub);
	    check_err(stat);
	    for(j=0;j<nrows;j++) {
	        stat = opendap_get_sequence_row(subnode,sub,subtype,j,&rowcolcontent);
	        printleaves1(subnode,rowcolcontent,subtype);
	    }
	}
	break;

    case OC_Grid:
    default: break;
    }

}

static void
dumpinstance(char* name, int rank, int* dims, size_t count, OCtype octype, char* memory)
{
    int i;
    char tmp[256];
    size_t delta = octypesize(octype);

    fprintf(stdout,"%s %s",octypetostring(octype),name);
    for(i=0;i<rank;i++) {
	fprintf(stdout,"[%u]",(unsigned int)dims[i]);
    }
    fprintf(stdout," =");
    for(i=0;i<count;i++) {
	fprintf(stdout," %s",octypeprint(octype,tmp,sizeof(tmp),memory+(i*delta)));
    }
    fprintf(stdout,"\n");
    fflush(stdout);
}
