/* Copyright 2009, UCAR/Unidata and OPeNDAP, Inc.
   See the COPYRIGHT file for more information. */

#undef TRACK

#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <assert.h>
#include <string.h>

#include "oc.h"
#include <oclist.h>
#include <ocbytes.h>

extern int ocdebug;

static OCerror stat;
static OCconnection gconn;

/* define a large stack of DUMPPATH instances */
struct DUMPPATH {
    size_t i,index,max;
    OCmode mode;
    char smode[1024];
    OCobject node;
    char name[1024];
    OCtype octype;
    OCtype etype;
} stack[2048];

static unsigned int stacknext;

static void
testpanic() {};

static void
pushstack(OCdata data)
{
    char* name = NULL;
    struct DUMPPATH* entry = stack+stacknext;
    if(oc_data_count(gconn,data,&entry->max)) testpanic();
    if(oc_data_index(gconn,data,&entry->index)) testpanic();
    if(oc_data_mode(gconn,data,&entry->mode)) testpanic();
    if(oc_data_object(gconn,data,&entry->node)) testpanic();
    if(oc_inq_name(gconn,entry->node,&name)) testpanic();
    strcpy(entry->name,name);
    free(name);
    if(oc_inq_class(gconn,entry->node,&entry->octype)) testpanic();
    if(oc_inq_primtype(gconn,entry->node,&entry->etype)) testpanic();
    switch (entry->mode) {
	case OCARRAYMODE: strcpy(entry->smode,"array"); break;
	case OCRECORDMODE:strcpy(entry->smode,"record"); break;
	case OCFIELDMODE:strcpy(entry->smode,"field"); break;
	case OCSCALARMODE:strcpy(entry->smode,"scalar"); break;
	default: break;
    }
   stacknext++;
}

static void popstack() {stacknext--;}

#ifdef TRACK
static void printstack(char* msg)
{
    size_t i;
    struct DUMPPATH* entry;
    fprintf(stderr,"\n%s@stack: %u\n",msg,stacknext);
    for(entry=stack,i=0;i<stacknext;i++,entry++) {
	fprintf(stderr,"    [%lu] (%s) %s[%lu/%lu]\n",
		(unsigned long)i,entry->smode,entry->name,
		(unsigned long)entry->index,
		(unsigned long)entry->max);
    }
}
#endif

#define CHECK(x) check_err((stat=(x)))

/* Command line arguments */
int     bracketed;	/* -B */
char*   constraint;	/* -C */
int     debug;		/* -d | -Dn */
int     dumpdatadds;    /* -X */
char*   fileurl;	/* -f | null */
int     generate;	/* -g */
int     inmemory;	/* -m */
int     optdas;		/* -p */
int     optdatadds;	/* -p */
int     optdds;		/* -p */
int     showattributes;	/* -M */
OClist* userparams;	/* -U */
int     verbose;	/* -v */


extern int optind;
extern int opterr;
extern char* optarg;

/* Forward*/
static void usage(char*);
static int fail(char*);
static void check_err(int stat);
static void dumpflags(void);

static OCerror processdata(void);
static OCerror printleaves(OCconnection, OCobject node, OCdata);
static OCerror printleavesr(OCconnection, OCobject node, OCdata,OCbytes*);
static OCerror dumpinstance(OCconnection, OCobject node, size_t count, void* memory, OCbytes*);
static char* generatedds(OCconnection,OCobject);
static OCerror generateddsr(OCconnection, OCobject, OCbytes*, int depth);
static char* generatedas(OCconnection,OCobject);
static OCerror generatedasr(OCconnection, OCobject, OCbytes*, int depth);
static OCerror generateddsattributes(OCconnection, OCobject node, OCbytes*, int);
static char* stringescape(char*);
static char* idescape(char*, char*, size_t);
static OCerror bracketdump(OCconnection, OCobject, size_t, void*, OCbytes*);

static void
init()
{
    bracketed	   = 0;
    constraint	   = NULL;
    debug	   = 0;
    dumpdatadds    = 0;
    fileurl	   = NULL;
    generate	   = 0;
    inmemory	   = 0;
    optdas	   = 0;
    optdatadds	   = 0;
    optdds	   = 0;
    showattributes = 0;
    userparams	   = oclistnew();
    verbose	   = 0;
}

int
main(int argc, char **argv)
{
    int c;

    init();

    opterr = 1;

    while ((c = getopt(argc, argv, "BC:dD:gmMp:U:vX")) != EOF) {
        switch (c) {
        case 'B': bracketed = 1; break;
	case 'C': constraint = strdup(optarg); break;
        case 'd': debug = 1; break;
        case 'D': debug = atoi(optarg); break;
	case 'f': fileurl = optarg; break;
        case 'g': generate = 1; break;
	case 'm': inmemory = 1; break;
	case 'M': showattributes = 1; break;
	case 'p':
	    if(strcasecmp(optarg,"das")==0) optdas=1;
	    else if(strcasecmp(optarg,"dds")==0) optdds=1;
	    else if(strcasecmp(optarg,"data")==0) optdatadds=1;
	    else if(strcasecmp(optarg,"datadds")==0) optdatadds=1;
	    else if(strcasecmp(optarg,"all")==0) {
		optdas = 1; optdds = 1; optdatadds = 1;
	    } else usage("unknown -p option");
	    break;
	case 'U': oclistpush(userparams,(ocelem)strdup(optarg)); break;
        case 'v': verbose = 1; break;
        case 'X': dumpdatadds = 1; break;
        default: usage("unknown option");
        }
    }

    if (debug > 0) {
        ocdebug = debug;
    }

    argc -= optind;
    argv += optind;

    if (argc > 0 && fileurl == NULL) {
        fileurl = strdup(argv[argc - 1]);
    }

    if (fileurl == NULL)
        fileurl = getenv("FILEURL");

    if (fileurl == NULL) {
        usage("no file url specified\n");
    }

    if (verbose)
        dumpflags();

    processdata();

    return 0;
}

static void
dumpflags(void)
{
    if(bracketed) fprintf(stderr," -B");
    if(constraint) fprintf(stderr," -C %s",constraint);
    if(debug) fprintf(stderr," -D%d",debug);
    if(generate) fprintf(stderr," -g");
    if(inmemory) fprintf(stderr," -m");
    if(optdas || optdds || optdatadds) {
	fprintf(stderr," -p");
	if(optdas) fprintf(stderr," das");
	if(optdds) fprintf(stderr," dds");
	if(optdatadds) fprintf(stderr," datadds");
    }
    if(showattributes) fprintf(stderr," -M");
    if(oclistlength(userparams) > 0) {
	unsigned int i;
	fprintf(stderr," -U");
	for(i=0;i<oclistlength(userparams);i++) {
	    fprintf(stderr," %s",(char*)oclistget(userparams,i));
	}
    }
    if(verbose) fprintf(stderr," -v");
    if(dumpdatadds) fprintf(stderr," -X");
    fprintf(stderr,"-f %s\n",fileurl);
}

static void
usage(char* msg)
{
    if(msg) fprintf(stderr,"error: %s\n",msg);
    fprintf(stderr,"usage: octest %s\n",
	"[-B] [-C <constraint>] [-d] [-D <level>] [-g] [-m] [-M] [-p das|dds|datadds] [-U <userparam>] [-v] <url>");
    fail(NULL);
}

static int
fail(char* msg)
{
    if(msg) fprintf(stderr,"fatalerror: %s\n",msg);
    fflush(stdout); fflush(stderr);
    exit(1);
}

/******************************************/
static void
check_err(OCerror stat)
{
    if(stat == OC_NOERR) return;
    fprintf(stderr,"error status returned: (%d) %s\n",stat,oc_errstring(stat));
    fail(NULL);
}

static OCerror
processdata(void)
{
    char* totalurl;
    unsigned int i;
    OCconnection conn;
    OCobject dasroot, ddsroot, dataroot;
    OCdata rootcontent;
    
    if(oclistlength(userparams) > 0) {
	OCbytes* buf = ocbytesnew();
	ocbytescat(buf,"[");
	for(i=0;i<oclistlength(userparams);i++) {
	    if(i>0) ocbytescat(buf,",");
	    ocbytescat(buf,(char*)oclistget(userparams,i));
	}
	ocbytescat(buf,"]");
	ocbytescat(buf,fileurl);
	totalurl = ocbytesdup(buf);
	ocbytesfree(buf);
    } else {
	totalurl = strdup(fileurl);
    }

    CHECK(oc_open(totalurl,&conn));
    free(totalurl);
    gconn = conn;

    /* Insert user parameters */
    for(i=0;i<oclistlength(userparams);i++) {
	char* parm = strdup((char*)oclistget(userparams,i));
	char* value = strchr(parm,'=');
	if(value == NULL) {value = "";} else {*value = '\0'; value++;}
        oc_clientparam_replace(conn,parm,value);
 	free(parm);
    }

    if(optdas) {
	CHECK(oc_fetch(conn,NULL,OCDAS,&dasroot));
        if(generate) {
	    char* das = generatedas(conn,dasroot);
	    fprintf(stdout,"%s",das);
	    free(das);
        } else {
	    const char* text = oc_inq_text(conn,dasroot);
	    fprintf(stdout,"%s",(text?text:"null"));
	}
    }
    if(optdds) {
	CHECK(oc_fetch(conn,NULL,OCDDS,&ddsroot));
	if(optdas) {
	    CHECK(oc_attach_das(conn,dasroot,ddsroot));
	}
        if(generate) {
	    char* dds = generatedds(conn,ddsroot);
	    fprintf(stdout,"%s",dds);
	    free(dds);
        } else {
	    const char* text = oc_inq_text(conn,ddsroot);
	    fprintf(stdout,"%s",(text?text:"null"));
	}
    }
    fflush(stdout);

    /* Read only if requested */
    if(!dumpdatadds && !optdatadds) goto done;

    stat = oc_fetch(conn,constraint,OCDATADDS,&dataroot);
    if(stat) {
	if(constraint) 
	    fprintf(stderr,"Cannot read DATADDS: %s?%s\n",fileurl,constraint);
	else
	    fprintf(stderr,"Cannot read DATADDS: %s\n",fileurl);
	exit(1);
    }

    if(dumpdatadds) {
        oc_dd(conn,dataroot);
        goto done;
    }

    if(inmemory) {
	CHECK(oc_compile(conn,dataroot));
    }

    /* Get the root content*/
    rootcontent = oc_data_new(conn);
    CHECK(oc_data_root(conn,dataroot,rootcontent));

    stacknext = 0;
    printleaves(conn,dataroot,rootcontent);

done:
    oc_close(conn);
    return OC_NOERR;
}

static OCerror
printleaves(OCconnection conn, OCobject node, OCdata content)
{
    OCerror stat;
    unsigned int i,nsubnodes;
    OCbytes* buffer;
    OCobject subnode;
    OCdata fieldcontent;

    buffer = ocbytesnew();
    fieldcontent = oc_data_new(conn);
    if(bracketed) {
	char* name;
	CHECK(oc_inq_nsubnodes(conn,node,&nsubnodes));
	for(i=0;i<nsubnodes;i++) {
	    CHECK(oc_inq_ith(conn,node,i,&subnode));
	    CHECK(oc_inq_name(conn,subnode,&name));
	    fprintf(stdout,"%s = ",name);
	    CHECK(oc_data_ith(conn,content,i,fieldcontent));
            printleavesr(conn,subnode,fieldcontent,buffer);
            fprintf(stdout,"%s\n",ocbytescontents(buffer));
	}
    } else {
        printleavesr(conn,node,content,buffer);
        fprintf(stdout,"%s",ocbytescontents(buffer));
    }
    oc_data_free(conn,fieldcontent);
    ocbytesfree(buffer);
    return OC_NOERR;
}

static OCerror
printleavesr(OCconnection conn, OCobject node, OCdata content, OCbytes* buffer)
{
    int stat;
    unsigned int nsubnodes;
    unsigned int i,rank;
    size_t count;
    OCtype octype,etype;
    OCmode mode;
    OCdata subdata  = oc_data_new(conn);

    CHECK(oc_inq_object(conn,node,NULL,
			&octype,&etype,NULL,&rank,&nsubnodes,NULL));
    CHECK(oc_data_mode(conn,content,&mode));

    switch (mode) {
    case OCFIELDMODE:
	    if(bracketed) ocbytescat(buffer,"{");
            for(i=0;i<nsubnodes;i++) {
	        OCobject subnode;
	        CHECK(oc_inq_ith(conn,node,i,&subnode));
	        CHECK(oc_data_ith(conn,content,i,subdata));
	        if(i > 0) if(bracketed) ocbytescat(buffer,", ");
		pushstack(content);
		printleavesr(conn,subnode,subdata,buffer);
   	        popstack();
	    }
	    if(bracketed) ocbytescat(buffer,"}");
	break;

    case OCARRAYMODE:
        {
	    /* Walk to each element and dump*/
	    CHECK(oc_data_count(conn,content,&count));
	    if(bracketed) ocbytescat(buffer,"{");
	    if(octype == OC_Primitive) {
		size_t memsize = oc_typesize(etype)*count;
	        char* memory = malloc(memsize);
		size_t offset = 0;
		size_t avail = memsize;
	        for(i=0;i<count;i++) {
	            CHECK(oc_data_get(conn,content,(void*)(memory+offset),avail,i,1));
		    offset += oc_typesize(etype);
		    avail -= oc_typesize(etype);
		}
		pushstack(content);
	        dumpinstance(conn,node,count,memory,buffer);
		popstack();
		free(memory);
	    } else { /* structure*/
	        for(i=0;i<count;i++) {
	            CHECK(oc_data_ith(conn,content,i,subdata));
		    if(i > 0) if(bracketed) ocbytescat(buffer,", ");
		    pushstack(content);
	            printleavesr(conn,node,subdata,buffer);
	            popstack();
		}
	    }
            if(bracketed) ocbytescat(buffer,"}");
	}
	break;

    case OCRECORDMODE:
	{
            for(i=0;;i++) {/* could have used countrecords*/
	        stat = oc_data_ith(conn,content,i,subdata);
		if(stat != OC_NOERR) {
		if(stat == OC_EINVALCOORDS) {
		    break; /* ith record does not exist*/
		}
	        CHECK(stat);
		}
		if(i > 0) if(bracketed) ocbytescat(buffer,", ");
  	        pushstack(content);
		printleavesr(conn,node,subdata,buffer);
	        popstack();
	    }
	}
	break;

    case OCSCALARMODE:
	{ /* singleton data item*/
	    void* memory;
	    memory = calloc(oc_typesize(etype),1);
	    pushstack(content);
	    CHECK(oc_data_get(conn,content,memory,oc_typesize(etype),0,1));
	    dumpinstance(conn,node,1,memory,buffer);
	    popstack();
	    free(memory);
	}
	break;

    default: break;
    }
    oc_data_free(conn,subdata);
    return OC_NOERR;
}

static char*
generatedds(OCconnection conn, OCobject root)
{
    unsigned int i;
    char* result;
    OCbytes* buffer = ocbytesnew();
    char* name;
    char tmp[1024];

    ocbytescat(buffer,"Dataset {\n");
    
    if(root != OCNULL) {
        unsigned int count;
	generateddsattributes(conn,root,buffer,2);
	CHECK(oc_inq_nsubnodes(conn,root,&count));
        for(i=0;i<count;i++) {
            OCobject node;
	    CHECK(oc_inq_ith(conn,root,i,&node));
	    generateddsr(conn,node,buffer,1);
	}
    }
    ocbytescat(buffer,"} ");
    CHECK(oc_inq_name(conn,root,&name));
    if(root != OCNULL) ocbytescat(buffer,idescape(name,tmp,sizeof(tmp)));
    free(name);
    ocbytescat(buffer,";\n");
    result = ocbytesdup(buffer);
    ocbytesfree(buffer);
    return result;    
}

static char blanks[2048];
#define BLANKSPERDENT 2

static char*
indent(int n)
{
    int nblanks = BLANKSPERDENT * n;
    memset(blanks,' ',nblanks);
    blanks[nblanks] = '\0';
    return blanks;
}

static OCerror
generateddsr(OCconnection conn, OCobject node, OCbytes* buffer, int depth)
{
    OCerror ocstatus = OC_NOERR;
    unsigned int i,rank,nattr,nsubnodes;
    OCtype octype, etype;
    OCobject container;
    OCobject* dimids;
    OCobject* subnodes;
    size_t size;
    char tmp[1024];
    char id1[1024];
    char* name;

    ocbytescat(buffer,indent(depth));

    /* get all info about the node */
    ocstatus = oc_inq_object(conn,node,&name,&octype,&etype,&container,
                             &rank,&nsubnodes,&nattr);
    ocstatus = oc_inq_subnodes(conn,node,&subnodes);

    switch (octype) {

    case OC_Primitive:
        snprintf(tmp,sizeof(tmp),"%s %s",
		oc_typetostring(etype),idescape(name,id1,sizeof(id1)));
	ocbytescat(buffer,tmp);
	/* dump dim info*/
	dimids = NULL;
	ocstatus = oc_inq_dimset(conn,node,&dimids);
	for(i=0;i<rank;i++) {
	    OCobject dim = dimids[i];
	    char* dimname = NULL;
	    CHECK(oc_inq_dim(conn,dim,&size,&dimname));
	    if(dimname == NULL)
	        snprintf(tmp,sizeof(tmp),"[%lu]",(unsigned long)size);
	    else
	        snprintf(tmp,sizeof(tmp),"[%s=%lu]",idescape(dimname,id1,sizeof(id1)),(unsigned long)size);
	    ocbytescat(buffer,tmp);
	    if(dimname) free(dimname);
	}
        ocbytescat(buffer,";\n");
	free(dimids);
	generateddsattributes(conn,node,buffer,depth+1);
	break;

    case OC_Sequence:
        snprintf(tmp,sizeof(tmp),"Sequence {\n");
        ocbytescat(buffer,tmp);
	for(i=0;i<nsubnodes;i++) {
	    OCobject subnode = subnodes[i];
	    generateddsr(conn,subnode,buffer,depth+1);
	}
	snprintf(tmp,sizeof(tmp),"%s} %s;\n",indent(depth),idescape(name,id1,sizeof(id1)));
        ocbytescat(buffer,tmp);
	generateddsattributes(conn,node,buffer,depth+1);
	break;

    case OC_Structure:
        snprintf(tmp,sizeof(tmp),"Structure {\n");
        ocbytescat(buffer,tmp);
	for(i=0;i<nsubnodes;i++) {
	    OCobject subnode = subnodes[i];
	    generateddsr(conn,subnode,buffer,depth+1);
	}
	snprintf(tmp,sizeof(tmp),"%s} %s",indent(depth),idescape(name,id1,sizeof(id1)));
	ocbytescat(buffer,tmp);
	dimids = NULL;
	ocstatus = oc_inq_dimset(conn,node,&dimids);
	for(i=0;i<rank;i++) {
	    char* dimname = NULL;
	    OCobject dim = dimids[i];
	    CHECK(oc_inq_dim(conn,dim,&size,&dimname));
	    if(dimname == NULL)
	        snprintf(tmp,sizeof(tmp),"[%lu]",(unsigned long)size);
	    else
	        snprintf(tmp,sizeof(tmp),"[%s=%lu]",dimname,(unsigned long)size);
	    ocbytescat(buffer,tmp);
	    if(dimname) free(dimname);
	}
        ocbytescat(buffer,";\n");
	free(dimids);
	generateddsattributes(conn,node,buffer,depth+1);
	break;

    case OC_Grid:
        snprintf(tmp,sizeof(tmp),"Grid {\n");
        ocbytescat(buffer,tmp);
	for(i=0;i<nsubnodes;i++) {
	    OCobject subnode = subnodes[i];
	    if(i <= 1) {
	        ocbytescat(buffer,indent(depth+1));
	        if(i==0) 
		    ocbytescat(buffer,"Array:\n");
	        else
		    ocbytescat(buffer,"Maps:\n");
	    }
	    generateddsr(conn,subnode,buffer,depth+2);
	}
	snprintf(tmp,sizeof(tmp),"%s} %s;\n",indent(depth),idescape(name,id1,sizeof(id1)));
        ocbytescat(buffer,tmp);
	generateddsattributes(conn,node,buffer,depth+1);
	break;

    default:
	snprintf(tmp,sizeof(tmp),"ocget DDS: unexpected type: %d",(int)octype);
	ocbytescat(buffer,tmp);
    }
    free(subnodes);
    if(name) free(name);
    return OC_NOERR;
}

static OCerror
generateddsattributes(OCconnection conn, OCobject node, OCbytes* buffer, int depth)
{
    unsigned int i,j;
    char tmp[128];
    unsigned int nattrs;
    char* aname;
    char* name;
    OCtype primtype;
    unsigned int nvalues;
    void* values;
    char id1[1024];

    CHECK(oc_inq_nattr(conn,node,&nattrs));
    CHECK(oc_inq_name(conn,node,&name));

    if(showattributes && nattrs > 0) {
	for(i=0;i<nattrs;i++) {
	    CHECK(oc_inq_attr(conn,node,i,&aname,&primtype,&nvalues,&values));
	    char* memory = (char*)values;
	    snprintf(tmp,sizeof(tmp),"%s%s %s:%s = ",indent(depth),
			oc_typetostring(primtype),idescape(name,id1,sizeof(id1)),aname);
	    ocbytescat(buffer,tmp);
	    for(j=0;j<nvalues;j++) {
		if(j > 0) ocbytescat(buffer,", ");
                CHECK(oc_typeprint(primtype,tmp,sizeof(tmp),(void*)memory));
		ocbytescat(buffer,tmp);
		memory += oc_typesize(primtype);
	    }
	    ocbytescat(buffer,";\n");
	    free(aname);
	    oc_attr_reclaim(primtype,nvalues,values);
	}
    }
    free(name);
    return OC_NOERR;
}

static char*
generatedas(OCconnection conn, OCobject root)
{
    unsigned int i, nsubnodes;
    char* result;
    OCbytes* buffer = ocbytesnew();

    CHECK(oc_inq_nsubnodes(conn,root,&nsubnodes));
    ocbytescat(buffer,"Attributes {\n");
    for(i=0;i<nsubnodes;i++) {
        OCobject node;
	CHECK(oc_inq_ith(conn,root,i,&node));
	generatedasr(conn,node,buffer,1);
    }
    ocbytescat(buffer,"}\n");
    result = ocbytesdup(buffer);
    ocbytesfree(buffer);
    return result;    
}

static OCerror
generatedasr(OCconnection conn, OCobject node, OCbytes* buffer, int depth)
{
    unsigned int i,nsubnodes,nvalues;
    char tmp[256];
    OCtype octype, etype;
    OCobject* subnodes = NULL;
    char* name = NULL;
    char id1[1024];

    /* get all info about the node */
    CHECK(oc_inq_object(conn,node,&name,&octype,&etype,NULL,
                             NULL,&nsubnodes,&nvalues));
    CHECK(oc_inq_subnodes(conn,node,&subnodes));

    if(octype == OC_Attributeset) {
	snprintf(tmp,sizeof(tmp),"%s%s {\n",indent(depth),idescape(name,id1,sizeof(id1)));
	ocbytescat(buffer,tmp);
	for(i=0;i<nsubnodes;i++) {
	    OCobject attr = subnodes[i];
	    generatedasr(conn,attr,buffer,depth+1);
	}
        ocbytescat(buffer,indent(depth));
        ocbytescat(buffer,"}\n");
    } else if(octype == OC_Attribute) {
	snprintf(tmp,sizeof(tmp),"%s%s %s",indent(depth),
		oc_typetostring(etype),idescape(name,id1,sizeof(id1)));
	ocbytescat(buffer,tmp);
	for(i=0;i<nvalues;i++) {
	    char* value;
	    OCtype ptype;
  	    CHECK(oc_inq_dasattr(conn,node,i,&ptype,&value));
            if(i > 0) ocbytescat(buffer,",");
	    if(ptype == OC_String || ptype == OC_URL) {
		char* se = stringescape(value);
	        snprintf(tmp,sizeof(tmp)," \"%s\"",se);
		free(se);
	    } else
	        snprintf(tmp,sizeof(tmp)," %s",value);
	    ocbytescat(buffer,tmp);
	    free(value);
	}
	ocbytescat(buffer,";\n");
    } else {
	snprintf(tmp,sizeof(tmp),"ocget DAS: unexpected type: %d",(int)octype);
	ocbytescat(buffer,tmp);
    }
    if(subnodes) free(subnodes);
    if(name) free(name);
    return OC_NOERR;
}

static char hexdigits[] = "0123456789abcdef";

/* Add escape characters to a string */
static char*
stringescape(char* s)
{
    size_t len;
    char* p;
    int c;
    char* escapedstring;

    if(s == NULL) return NULL;
    len = strlen(s);
    escapedstring = (char*)malloc(4*len);
    p = escapedstring;
    while((c=*s++)) {
	if(c == '"' || c == '\\') {*p++ = '\\'; *p++ = c;}
	else if (c < ' ' || c >= 127) {
	    *p++ = '\\'; *p++ = 'x';
	    *p++ = hexdigits[(c & 0xf0)>>4];
	    *p++ = hexdigits[(c & 0xf)];
	} else
	    *p++ = c;
    }
    *p = '\0';
    return escapedstring;
}

static char idchars[] = "_.";

/* Add escape characters to an identifier */
static char*
idescape(char* id, char* escapeid, size_t esize)
{
    size_t len;
    char* p;
    int c;

    if(id == NULL) return NULL;
    len = strlen(id);
    p = escapeid;
    *p = '\0';
    esize--; /* leave room for null */
    while(esize-- > 0 && (c=*id++)) {
	if(c >= '0' && c <= '9') {*p++ = c;}
	else if(c >= 'a' && c <= 'z') {*p++ = c;}
	else if(c >= 'A' && c <= 'Z') {*p++ = c;}
	else if(strchr(idchars,c) != NULL) {*p++ = c;}
	else {
	    *p++ = '%';
	    *p++ = hexdigits[(c & 0xf0)>>4];
	    *p++ = hexdigits[(c & 0xf)];
	}
    }
    *p = '\0';
    return escapeid;
}

#ifdef IGNORE
static OCerror
dumpinactive(OCconnection conn, OCobject node, OCbytes* buffer)
{
    unsigned int i,rank;
    OCtype etype;
    OCtype octype;
    OCbytes* path;
    char tmp[1024];
    char id[1024];
    struct DUMPPATH* entry;
   
    CHECK(oc_inq_primtype(conn,node,&etype));
    CHECK(oc_inq_class(conn,node,&octype));

    /* construct the name from the stack: do not duplicate nodes */
    path = ocbytesnew();
    for(entry=stack,i=0;i<stacknext;i++,entry++) {
	if(entry->octype == OC_Dataset) continue;
	(void)idescape(entry->name,id,sizeof(id));
	switch (entry->mode) {
	case OCARRAYMODE:
	    ocbytescat(path,"/");
	    ocbytescat(path,id);
	    if(entry->octype != OC_Primitive) {
	        sprintf(tmp,"[%lu]",(unsigned long)entry->index);
	        ocbytescat(path,tmp);
	    }
	    break;
	case OCRECORDMODE:
	    ocbytescat(path,"/");
	    ocbytescat(path,id);
	    sprintf(tmp,"[%lu]",(unsigned long)entry->index);
	    ocbytescat(path,tmp);
	    break;
	case OCFIELDMODE:
	case OCSCALARMODE:
	    /* Only include if the previous entry was not same node */
	    if(stack[i-1].node != entry->node) {
	        ocbytescat(path,"/");
	        ocbytescat(path,id);
	    }
	    break;
	default: ocbytescat(path,"?"); break;
 	}
    }
    /* work with the final entry */
    entry = stack + (stacknext - 1);
    CHECK(oc_inq_rank(conn,entry->node,&rank));
    if(rank == 0) {
        snprintf(tmp,sizeof(tmp),"%s %s",
		oc_typetostring(entry->etype),
		ocbytescontents(path));
    } else {
        snprintf(tmp,sizeof(tmp),"%s %s[%s%lu]",
		oc_typetostring(entry->etype),
		ocbytescontents(path),
		(entry->max == 0)?"":"0..",
		(entry->max == 0)?entry->index:(entry->max-1));
    }
    ocbytescat(buffer,tmp);
    ocbytescat(buffer,"\n");
    ocbytesfree(path);
    return OC_NOERR;
}
#endif

static OCerror
dumpinstance(OCconnection conn, OCobject node, size_t count, void* memory, OCbytes* buffer)
{
    unsigned int i,rank;
    size_t delta;
    OCtype etype;
    OCtype octype;
    OCbytes* path;
    size_t max;
    char tmp[1024];
    char id[1024];
    struct DUMPPATH* entry;
   
    if(bracketed) {return bracketdump(conn,node,count,memory,buffer);}

    CHECK(oc_inq_primtype(conn,node,&etype));
    CHECK(oc_inq_class(conn,node,&octype));
    delta = oc_typesize(etype);
    max = count - 1;

#ifdef TRACK
    printstack("dumpinstance");
#endif

    /* construct the name from the stack: do not duplicate nodes */

    path = ocbytesnew();
    for(entry=stack,i=0;i<stacknext;i++,entry++) {
	if(entry->octype == OC_Dataset) continue;
	(void)idescape(entry->name,id,sizeof(id));
	switch (entry->mode) {
	case OCARRAYMODE:
	    ocbytescat(path,"/");
	    ocbytescat(path,id);
	    if(entry->octype != OC_Primitive) {
	        sprintf(tmp,"[%lu]",(unsigned long)entry->index);
	        ocbytescat(path,tmp);
	    }
	    break;
	case OCRECORDMODE:
	    ocbytescat(path,"/");
	    ocbytescat(path,id);
	    sprintf(tmp,"[%lu]",(unsigned long)entry->index);
	    ocbytescat(path,tmp);
	    break;
	case OCFIELDMODE:
	case OCSCALARMODE:
	    /* Only include if the previous entry was not same node */
	    if(stack[i-1].node != entry->node) {
	        ocbytescat(path,"/");
	        ocbytescat(path,id);
	    }
	    break;
	default: ocbytescat(path,"?"); break;
 	}
    }
    /* work with the final entry */
    entry = stack + (stacknext - 1);
    assert(entry->octype == OC_Primitive);
    CHECK(oc_inq_rank(conn,entry->node,&rank));
    if(rank == 0) {
        snprintf(tmp,sizeof(tmp),"%s %s",
		oc_typetostring(entry->etype),
		ocbytescontents(path));
    } else {
        snprintf(tmp,sizeof(tmp),"%s %s[%s%lu]",
		oc_typetostring(entry->etype),
		ocbytescontents(path),
		(entry->max == 0)?"":"0..",
		(entry->max == 0)?entry->index:(entry->max-1));
    }
    ocbytescat(buffer,tmp);

    for(i=0;i<count;i++) {
	ocbytescat(buffer," ");	
	oc_typeprint(etype,tmp,sizeof(tmp),(memory+(i*delta)));
	ocbytescat(buffer,tmp);	
    }
    ocbytescat(buffer,"\n");
    ocbytesfree(path);
    /* If the type was string, then free it up */
    if(etype == OC_String || etype == OC_URL) {
	char** strings = (char**)memory;
	for(i=0;i<count;i++) free(strings[i]);
    }
    return OC_NOERR;
}

static OCerror
bracketdump(OCconnection conn, OCobject node, size_t count, void* memory, OCbytes* buffer)
{
    unsigned int i;
    size_t delta;
    OCtype etype;
    OCtype octype;
    size_t max;
    char tmp[1024];
   
#ifdef TRACK
    printstack("bracketdump");
#endif

    CHECK(oc_inq_primtype(conn,node,&etype));
    CHECK(oc_inq_class(conn,node,&octype));
    delta = oc_typesize(etype);
    max = count - 1;

    for(i=0;i<count;i++) {
	if(i > 0) ocbytescat(buffer,", ");
	else ocbytescat(buffer," ");	
	oc_typeprint(etype,tmp,sizeof(tmp),(memory+(i*delta)));
	ocbytescat(buffer,tmp);	
    }

    /* If the type was string, then free it up */
    if(etype == OC_String || etype == OC_URL) {
	char** strings = (char**)memory;
	for(i=0;i<count;i++) free(strings[i]);
    }
    return OC_NOERR;
}
